const std = @import("std");
const mem = std.mem;
const heap = std.heap;

const c = @import("libvirt-c.zig").c;
const h = @import("libvirt-helper.zig");
const err = @import("libvirt-error.zig");
const Connection = @import("libvirt-connection.zig");
const Volume = @import("libvirt-volume.zig");
const VirError = err.VirError;
const Pool = @This();

ptr: c.virStoragePoolPtr,
arena: *heap.ArenaAllocator,

pub fn init(ptr: c.virStoragePoolPtr, allocator: mem.Allocator) Pool {
    var arena = heap.ArenaAllocator.init(allocator);
    return .{
        .ptr = ptr,
        .arena = &arena,
    };
}
pub fn deinit(self: *const Pool) void {
    self.arena.deinit();
}

// pub fn findStoragePoolSources(
//     conn: *const Connection,
//     type: []const u8,
//     srcSpec: []const u8,
//     flags: []const Flags,
// ) VirError![]u8 {} // TODO
// pub fn getStoragePoolCapabilities(
//     conn: *const Connection,
//     flags: []const Flags,
// ) VirError![]u8 {} // TODO
pub fn listAllStoragePools(
    conn: *const Connection,
    // pools: [*]c.virStoragePoolPtr,
    flags: []const ListFlags,
) VirError![]c.virStoragePoolPtr {
    var list: [*]c.virStoragePoolPtr = undefined;
    const num = c.virConnectListAllStoragePools(
        conn.ptr,
        @ptrCast(&list),
        h.intFromFlags(ListFlags, flags),
    );
    return if (num < 0) err.handleError() else list[0..@intCast(num)];
}
// pub fn listDefinedStoragePools(
//     conn: *const Connection,
//     names: [*]const []const u8,
//     maxnames: u32,
// ) VirError!u32 {} // TODO
// pub fn listStoragePools(
//     conn: *const Connection,
//     names: [*]const []const u8,
//     maxnames: u32,
// ) VirError!u32 {} // TODO
// pub fn numOfDefinedStoragePools(conn: *const Connection) VirError!u32 {}
// pub fn numOfStoragePools(conn: *const Connection) VirError!u32 {}
// TODO event handling callbacks

pub fn build(self: *const Pool, flags: []const BuildFlags) VirError!void {
    if (c.virStoragePoolBuild(self.ptr, h.intFromFlags(BuildFlags, flags)) < 0) return err.handleError();
}
// pub fn create(conn: *const Connection) VirError!void {} // TODO
// pub fn createXML(conn: *const Connection) VirError!void {} // TODO
// pub fn defineXML(conn: *const Connection) VirError!void {} // TODO
pub fn delete(self: *const Pool, flags: []const DeleteFlags) VirError!void {
    if (c.virStoragePoolDelete(self.ptr, h.intFromFlags(DeleteFlags, flags)) < 0) return err.handleError();
}
pub fn destroy(self: *const Pool) VirError!void {
    if (c.virStoragePoolDestroy(self.ptr) < 0) return err.handleError();
}
pub fn free(self: *const Pool) VirError!void {
    if (c.virStoragePoolFree(self.ptr) < 0) return err.handleError();
}

pub fn getAutostart(self: *const Pool, autostart: *c_int) VirError!void {
    if (c.virStoragePoolGetAutostart(self.ptr, autostart) < 0) return err.handleError();
}
pub fn getConnect(self: *const Pool) VirError!c.virConnectPtr {
    return if (c.virStoragePoolGetConnect(self.ptr)) |ptr| ptr else err.handleError();
}
pub fn getInfo(self: *const Pool, info: c.virStoragePoolInfoPtr) VirError!void {
    if (c.virStoragePoolGetInfo(self.ptr, info) < 0) return err.handleError();
}
pub fn getName(self: *const Pool) VirError![]const u8 {
    const name = c.virStoragePoolGetName(self.ptr);
    const str = h.string(name);
    return if (str.len == 0) err.handleError() else str;
}
// pub fn getUUID(self: *const Pool) VirError!void {} // TODO
// pub fn getUUIDString(self: *const Pool) VirError!void {} // TODO
pub fn getXMLDesc(self: *const Pool, flags: []const XMLFlags) VirError![]const u8 {
    const desc = c.virStoragePoolGetXMLDesc(self.ptr, h.intFromFlags(XMLFlags, flags));
    const str = h.string(desc);
    return if (str.len == 0) err.handleError() else str;
}

pub fn isActive(self: *const Pool) VirError!bool {
    return switch (c.virStoragePoolIsActive(self.ptr)) {
        0 => false,
        1 => true,
        else => err.handleError(),
    };
}
pub fn isPersistent(self: *const Pool) VirError!bool {
    return switch (c.virStoragePoolIsPersistent(self.ptr)) {
        0 => false,
        1 => true,
        else => err.handleError(),
    };
}
// pub fn listAllVolumes(self: *const Pool) void {} // TODO
// pub fn listVolumes(self: *const Pool) void {} // TODO

pub fn lookupByName(conn: *const Connection, name: []const u8) VirError!Pool {
    if (c.vir(conn.ptr, name)) |p| p else err.handleError();
}
pub fn lookupByTargetPath(conn: *const Connection, path: []const u8) VirError!Pool {
    if (c.vir(conn.ptr, path)) |p| p else err.handleError();
}
pub fn lookupByUUID(conn: *const Connection, uuid: []const u8) VirError!Pool {
    if (c.vir(conn.ptr, uuid)) |p| p else err.handleError();
}
pub fn lookupByUUIDString(conn: *const Connection, uuid: []const u8) VirError!Pool {
    if (c.vir(conn.ptr, uuid)) |p| p else err.handleError();
}
pub fn lookupByVolume(vol: *const Volume) VirError!Pool {
    if (c.vir(vol.ptr)) |p| p else err.handleError();
}

pub fn numOfVolumes(self: *const Pool) VirError!u32 {
    const num = c.virStoragePoolNumOfVolumes(self.ptr);
    return if (num < 0) err.handleError() else num;
}
pub fn ref(self: *const Pool) VirError!void {
    if (c.virStoragePoolRef(self.ptr) < 0) return err.handleError();
}
pub fn refresh(self: *const Pool, flags: []const RefreshFlags) VirError!void {
    if (c.virStoragePoolRefresh(self.ptr, h.intFromFlags(RefreshFlags, flags)) < 0) return err.handleError();
}
pub fn setAutostart(self: *const Pool, autostart: bool) VirError!void {
    const i = if (autostart) 1 else 0;
    if (c.virStoragePoolSetAutostart(self.ptr, i) < 0) return err.handleError();
}
pub fn undefine(self: *const Pool) VirError!void {
    if (c.virStoragePoolUndefine(self.ptr) < 0) return err.handleError();
}

// pub fn createVolume(self: *const Pool, xml: String, flags: []const Volume.CreateFlags) !Volume {
//     const volume = c.virStorageVolCreateXML(self.ptr, @ptrCast(xml.str), intFromFlags(Volume.CreateFlags, flags));
//     return if (volume) |v| Volume.init(v, self.arena.allocator()) else err.handleError();
// }

// pub fn lookupVolumeByName(self: *const Pool, name: []const u8) !Volume {
//     const volume = c.virStorageVolLookupByName(self.ptr, @ptrCast(name));
//     return if (volume) |v| Volume.init(v, self.arena.allocator()) else err.handleError();
// }

// pub fn iterateVolumes(self: *const Pool, flags: []const ListFlags) Volume.VolumeIterator {
//     return Volume.VolumeIterator.init(
//         c.virStorageVolumePtr,
//         Pool,
//         Volume.ListFlags,
//         self,
//         flags,
//         self.arena.allocator(),
//         c.virStoragePoolListAllVolumes,
//     );
// }

pub fn lookupVolumeByKey(self: *const Pool, key: []const u8) VirError!Volume {
    const volume = Volume.lookupByKey(self, key);
    return if (volume) |v| v else err.handleError();
}
pub fn lookupVolumeByName(self: *const Pool, name: []const u8) VirError!Volume {
    const volume = Volume.lookupByName(self, name);
    return if (volume) |v| v else err.handleError();
}
pub fn lookupVolumeByPath(self: *const Pool, path: []const u8) VirError!Volume {
    const volume = Volume.lookupByPath(self, path);
    return if (volume) |v| v else err.handleError();
}

pub const ListFlags = enum(c_uint) {
    Inactive = c.VIR_CONNECT_LIST_STORAGE_POOLS_INACTIVE,
    Active = c.VIR_CONNECT_LIST_STORAGE_POOLS_ACTIVE,
    Persistent = c.VIR_CONNECT_LIST_STORAGE_POOLS_PERSISTENT,
    Transient = c.VIR_CONNECT_LIST_STORAGE_POOLS_TRANSIENT,
    Autostart = c.VIR_CONNECT_LIST_STORAGE_POOLS_AUTOSTART,
    NoAutostart = c.VIR_CONNECT_LIST_STORAGE_POOLS_NO_AUTOSTART,
    Dir = c.VIR_CONNECT_LIST_STORAGE_POOLS_DIR,
    Fs = c.VIR_CONNECT_LIST_STORAGE_POOLS_FS,
    NetFs = c.VIR_CONNECT_LIST_STORAGE_POOLS_NETFS,
    Logical = c.VIR_CONNECT_LIST_STORAGE_POOLS_LOGICAL,
    Disk = c.VIR_CONNECT_LIST_STORAGE_POOLS_DISK,
    Iscsi = c.VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI,
    Scsi = c.VIR_CONNECT_LIST_STORAGE_POOLS_SCSI,
    Mpath = c.VIR_CONNECT_LIST_STORAGE_POOLS_MPATH,
    Rbd = c.VIR_CONNECT_LIST_STORAGE_POOLS_RBD,
    Sheepdog = c.VIR_CONNECT_LIST_STORAGE_POOLS_SHEEPDOG,
    Gluster = c.VIR_CONNECT_LIST_STORAGE_POOLS_GLUSTER,
    Zfs = c.VIR_CONNECT_LIST_STORAGE_POOLS_ZFS,
    Vstorage = c.VIR_CONNECT_LIST_STORAGE_POOLS_VSTORAGE,
    IscsiDirect = c.VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI_DIRECT,
};
pub const BuildFlags = enum(c_uint) {
    New = c.VIR_STORAGE_POOL_BUILD_NEW,
    Repair = c.VIR_STORAGE_POOL_BUILD_REPAIR,
    Resize = c.VIR_STORAGE_POOL_BUILD_RESIZE,
    NoOverwrite = c.VIR_STORAGE_POOL_BUILD_NO_OVERWRITE,
    Overwrite = c.VIR_STORAGE_POOL_BUILD_OVERWRITE,
};
pub const CreateFlags = enum(c_uint) {
    Normal = c.VIR_STORAGE_POOL_CREATE_NORMAL,
    WithBuild = c.VIR_STORAGE_POOL_CREATEWITH_BUILD_,
    WithBuildOverwrite = c.VIR_STORAGE_POOL_CREATE_WITH_BUILD_OVERWRITE,
    WithBuildNoOverwrite = c.VIR_STORAGE_POOL_CREATE_WITH_BUILD_NO_OVERWRITE,
};
pub const DefineFlags = enum(c_uint) {
    Validate = c.VIR_STORAGE_POOL_DEFINE_VALIDATE,
};
pub const DeleteFlags = enum(c_uint) {
    Normal = c.VIR_STORAGE_POOL_DELETE_NORMAL,
    Zeroed = c.VIR_STORAGE_POOL_DELETE_ZEROED,
};
pub const RefreshFlags = enum(c_uint) {};
pub const XMLFlags = enum(c_uint) {
    Inactive = c.VIR_STORAGE_XML_INACTIVE,
};

pub const Info = struct {
    state: State,
    capacity: u64,
    allocation: u64,
    available: u64,
};

pub const State = enum(c_uint) {
    Inactive = c.VIR_STORAGE_POOL_INACTIVE,
    Building = c.VIR_STORAGE_POOL_BUILDING,
    Running = c.VIR_STORAGE_POOL_RUNNING,
    Degraded = c.VIR_STORAGE_POOL_DEGRADED,
    Inaccessible = c.VIR_STORAGE_POOL_INACCESSIBLE,
    Last = c.VIR_STORAGE_POOL_STATE_LAST,
};
pub const EventID = enum(c_uint) {
    Lifecycle = c.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE,
    Refresh = c.VIR_STORAGE_POOL_EVENT_ID_REFRESH,
    Last = c.VIR_STORAGE_POOL_EVENT_ID_LAST,
};
pub const EventLifecycleType = enum(c_uint) {
    Defined = c.VIR_STORAGE_POOL_EVENT_DEFINED,
    Undefined = c.VIR_STORAGE_POOL_EVENT_UNDEFINED,
    Started = c.VIR_STORAGE_POOL_EVENT_STARTED,
    Stopped = c.VIR_STORAGE_POOL_EVENT_STOPPED,
    Created = c.VIR_STORAGE_POOL_EVENT_CREATED,
    Deleted = c.VIR_STORAGE_POOL_EVENT_DELETED,
    Last = c.VIR_STORAGE_POOL_EVENT_LAST,
};

// pub const PoolIterator = Iterator(Pool, c.virStoragePoolPtr, c.virStoragePoolFree);
