const std = @import("std");
const mem = std.mem;
const heap = std.heap;

const h = @import("libvirt-helper.zig");
const c = @import("libvirt-c.zig").c;
const err = @import("libvirt-error.zig");
const Connection = @import("libvirt-connection.zig");
const Pool = @import("libvirt-pool.zig");
const VirError = err.VirError;
const Volume = @This();

ptr: c.virStorageVolPtr,
arena: *heap.ArenaAllocator,

pub fn init(ptr: c.virStorageVolPtr, allocator: mem.Allocator) Volume {
    var arena = heap.ArenaAllocator.init(allocator);
    return .{
        .ptr = ptr,
        .arena = &arena,
    };
}
pub fn deinit(self: *const Volume) void {
    self.arena.deinit();
}

// pub fn createXML(
//     pool: *const Pool,
//     xml: h.String,
//     flags: []const CreateFlags,
// ) void {} // TODO
// pub fn createXMLFrom(
//     pool: *const Pool,
//     xml: h.String,
//     flags: []const CreateFlags,
// ) void {} // TODO
pub fn delete(self: *const Volume, flags: []const DeleteFlags) VirError!void {
    if (c.virStorageVolDelete(self.ptr, h.intFromFlags(DeleteFlags, flags)) < 0)
        return err.handleError();
}
// pub fn download(self: *const Volume) VirError!void {} // TODO
pub fn free(self: *const Volume) VirError!void {
    if (c.virStorageVolFree(self.ptr) < 0) return err.handleError();
}

pub fn getConnect(self: *const Volume) VirError!c.virConnectPtr {
    return if (c.virStorageVolGetConnect(self.ptr)) |conn| conn else err.handleError();
}
pub const GetInfoError = mem.Allocator.Error || VirError;
pub fn getInfo(self: *const Volume) GetInfoError!*Info {
    var info = try self.arena.allocator().create(Info);
    errdefer self.arena.allocator().destroy(info);
    const p: c.virStorageVolInfoPtr = undefined;
    if (c.virStorageVolGetInfo(self.ptr, p) < 0) return err.handleError();
    info.type = p.*.type;
    info.capacity = p.*.capacity;
    info.allocation = p.*.allocation;
    return info;
} // TODO
// pub fn getInfoFlags(self: *const Volume, flags: []const InfoFlags) GetInfoError!void {} // TODO
pub fn getKey(self: *const Volume) VirError![]const u8 {
    return if (c.virStorageVolGetKey(self.ptr)) |v| v else err.handleError();
}
pub fn getName(self: *const Volume) VirError![]const u8 {
    return if (c.virStorageVolGetName(self.ptr)) |v| v else err.handleError();
}
pub fn getPath(self: *const Volume) VirError![]const u8 {
    // TODO use h.String?
    return if (c.virStorageVolGetPath(self.ptr)) |v| v else err.handleError();
}
// pub fn getXMLDesc(self: *const Volume) void {} // TODO

pub fn lookupByKey(conn: *const Connection, key: []const u8) VirError!Volume {
    return if (c.virStorageVolLookupByKey(conn.ptr, key)) |v| v else err.handleError();
}
pub fn lookupByName(conn: *const Connection, name: []const u8) VirError!Volume {
    return if (c.virStorageVolLookupByName(conn.ptr, name)) |v| v else err.handleError();
}
pub fn lookupByPath(conn: *const Connection, path: []const u8) VirError!Volume {
    return if (c.virStorageVolLookupByPath(conn.ptr, path)) |v| v else err.handleError();
}

pub fn ref(self: *const Volume) VirError!void {
    if (c.virStorageVolRef(self.ptr) < 0) return err.handleError();
}
pub fn resize(self: *const Volume, capacity: u64, flags: []const ResizeFlags) VirError!void {
    if (c.virStorageVolResize(self.ptr, capacity, h.intFromFlags(ResizeFlags, flags)) < 0) return err.handleError();
}
// pub fn upload(self: *const Volume) void {} // TODO
pub fn wipe(self: *const Volume, flags: []const WipeFlags) VirError!void {
    if (c.virStorageVolWipe(self.ptr, h.intFromFlags(WipeFlags, flags)) < 0) return err.handleError();
}
pub fn wipePattern(self: *const Volume, algorithm: WipeAlgorithm, flags: []const WipeFlags) VirError!void {
    if (c.virStorageVolWipePattern(
        self.ptr,
        @intFromEnum(algorithm),
        h.intFromFlags(WipeFlags, flags),
    ) < 0) return err.handleError();
}

pub const ListFlags = enum(c_uint) {};
pub const CreateFlags = enum(c_uint) {
    PreallocMetadata = c.VIR_STORAGE_VOL_CREATE_PREALLOC_METADATA,
    Reflink = c.VIR_STORAGE_VOL_CREATE_REFLINK,
    Validate = c.VIR_STORAGE_VOL_CREATE_VALIDATE,
};
pub const DeleteFlags = enum(c_uint) {
    Normal = c.VIR_STORAGE_VOL_DELETE_NORMAL,
    Zeroed = c.VIR_STORAGE_VOL_DELETE_ZEROED,
    WithSnapshots = c.VIR_STORAGE_VOL_DELETE_WITH_SNAPSHOTS,
};
pub const DownloadFlags = enum(c_uint) {
    SparseStream = c.VIR_STORAGE_VOL_DOWNLOAD_SPARSE_STREAM,
};
pub const InfoFlags = enum(c_uint) {
    UseAllocation = c.VIR_STORAGE_VOL_USE_ALLOCATION,
    GetPhysical = c.VIR_STORAGE_VOL_GET_PHYSICAL,
};
pub const ResizeFlags = enum(c_uint) {
    Allocate = c.VIR_STORAGE_VOL_RESIZE_ALLOCATE,
    Delta = c.VIR_STORAGE_VOL_RESIZE_DELTA,
    Shrink = c.VIR_STORAGE_VOL_RESIZE_SHRINK,
};
pub const UploadFlags = enum(c_uint) {};
pub const WipeFlags = enum(c_uint) {};
pub const Info = struct {
    type: Type,
    capacity: u64,
    allocation: u64,
};
pub const Type = enum(c_uint) {
    File = c.VIR_STORAGE_VOL_FILE,
    Block = c.VIR_STORAGE_VOL_BLOCK,
    Dir = c.VIR_STORAGE_VOL_DIR,
    Network = c.VIR_STORAGE_VOL_NETWORK,
    Netdir = c.VIR_STORAGE_VOL_NETDIR,
    Ploop = c.VIR_STORAGE_VOL_PLOOP,
    Last = c.VIR_STORAGE_VOL_LAST,
};
pub const WipeAlgorithm = enum(c_uint) {
    Zero = c.VIR_STORAGE_VOL_WIPE_ALG_ZERO,
    Nnsa = c.VIR_STORAGE_VOL_WIPE_ALG_NNSA,
    Dod = c.VIR_STORAGE_VOL_WIPE_ALG_DOD,
    Bsi = c.VIR_STORAGE_VOL_WIPE_ALG_BSI,
    Gutmann = c.VIR_STORAGE_VOL_WIPE_ALG_GUTMANN,
    Schneier = c.VIR_STORAGE_VOL_WIPE_ALG_SHNEIER,
    Pfitzner7 = c.VIR_STORAGE_VOL_WIPE_ALG_PFITZNER7,
    Pfitzner33 = c.VIR_STORAGE_VOL_WIPE_ALG_PFITZNER33,
    Random = c.VIR_STORAGE_VOL_WIPE_ALG_RANDOM,
    Trim = c.VIR_STORAGE_VOL_WIPE_ALG_TRIM,
    Last = c.VIR_STORAGE_VOL_WIPE_ALG_LAST,
};

// pub const VolumeIterator = Iterator(Volume, c.virStorageVolPtr, c.virStorageVolFree);
